# Version file autogenerated by the build process
version = '0.11.1'
build = 1132
vcs_info = 'Git SHA1 bddd6219 committed on 2023-05-13'
package_name = 'ksconf'

if __name__ == '__main__':
    print('KSCONF_VERSION="0.11.1"\nKSCONF_BUILD="1132"\nKSCONF_VCS_INFO="Git SHA1 bddd6219 committed on 2023-05-13"')
